// main.cpp
//
// ICS 65 Fall 2012
// Code Example
//
// This program generates sequences of random numbers according to two
// different distributions: a uniform_int_distribution and a custom
// distribution I've written called a NormalIntDistribution, which uses
// a normal_distribution to generate doubles, but then truncates them to
// integer values (by using casting) and requires them to be in a range
// between a minimum and maximum.  In either case, the number of
// occurrences of each value are stored in a map and then printed to the
// console.
//
// I'm not using "using namespace std" or "using std::xxxxxx" here,
// because I want it to be clear, when reading this code, which things
// are part of the standard library and which are not.

#include <iostream>
#include <iomanip>
#include <random>
#include <ctime>
#include <map>
#include <string>
#include "NormalIntDistribution.h"



// This function prints the results generated by one distribution test.
// A couple of things of note here:
//
// (1) The "range-based for loop", when used on a map, returns a sequence
//     of std::pair<K, V> (where K is the type of key in the map and V
//     is the type of value).  These are returned in ascending order of
//     the keys.
//
// (2) I'm using what are called "I/O manipulators" like std::setw() and
//     std::right to control the formatting of the output.  As I don't use
//     these often, I find I often have to look them up in order to remember
//     how to use them, but cppreference.com is a great place to find out
//     more.

void printResults(const std::map<unsigned int, unsigned int>& results, const std::string& testName)
{
	std::cout << "Results - " << testName << std::endl;

	if (results.size() > 0)
	{
		for (auto p : results)
		{
			std::cout << std::setw(8) << std::right << p.first;
			std::cout << ": ";
			std::cout << p.second << std::endl;
		}
	}

	std::cout << std::endl;
}



// This function takes two parameters, a random number engine and a
// distribution, and runs a "distribution test" on them, by doing
// the following:
//
// * Generating 100,000 random numbers from the distribution.
// * Stores the number of occurrences of each random number into a
//   map, so we know how many times each number was chosen.
//
// You might be wondering why this is a template function.  That's because
// the random number engines and distributions don't inherit from a
// common base class; instead, what makes them similar is that they have
// a similar set of member functions.  So the only way to write a function
// that can take any random number engine and any distribution as parameters
// is to write it as a template function.  Any time I call this function
// with a new combination of engine type and distribution type, this
// template will be used by the compiler to automatically generate a new
// version of the function for those types.
//
// This function does make certain assumptions about the engine's type
// and the distribution's type:
//
// * The distribution can be called like a function (i.e., it overloads
//   the () operator) that takes an engine as a parameter
// * The result that comes back from that function is an unsigned int
//   (or, at least, is *compatible* with an unsigned int)
//
// Attempting to use this function with engines or distributions that
// do not follow these rules will cause errors at compile time.

template <typename EngineType, typename DistributionType>
std::map<unsigned int, unsigned int> runDistributionTest(EngineType& engine, DistributionType& distribution)
{
	std::map<unsigned int, unsigned int> results;

	for (int i = 0; i < 100000; i++)
	{
		unsigned int number = distribution(engine);

		if (results.count(number) == 0)
		{
			results[number] = 1;
		}
		else
		{
			results[number]++;
		}
	}

	return results;
}



int main()
{
	// First, we create a random number engine and seed it using the current time.
	std::mt19937 engine;
	engine.seed(static_cast<unsigned long>(std::time(nullptr)));

	// Now, we create two different distributions that we want to use: a
	// uniform_int_distribution (part of the standard library) and a
	// NormalIntDistribution (a custom distribution that's part of this
	// example).
	std::uniform_int_distribution<unsigned int> uniform(1, 100);
	NormalIntDistribution<unsigned int> normalInt(50, 10, 1, 100);

	// Finally, we run tests of each distribution and print their results
	// to the console.  We should expect the uniform distribution to choose
	// each possible result a reasonably similar number of times, while we
	// expect the NormalIntDistribution to choose numbers closer to the
	// mean (50) more often than numbers further away.
	printResults(runDistributionTest(engine, uniform), "Uniform from 1..100");
	printResults(runDistributionTest(engine, normalInt), "NormalIntDistribution from 1..100");

	return 0;
}
